"""pyslabs info module
"""

name            = "pyslabs"
version         = "0.1.3"
description     = "Pure Python Paralle I/O"
long_description = """Pure Python Parallel I/O that supports large datasets of
 all pickle-able objects and well-known data frameworks"""
author          = "Youngsung Kim"
